#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <getopt.h>
#include "unittools.h"

#define  CONVERT         8.210
#define  SRCUNITNAME     "stadions"
#define  SRCUNIT         "st"
#define  DSTUNITNAME     "miles"
#define  DSTUNITNAMESING "mile"
#define  DSTUNIT         "mi"

typedef unsigned char uc;

void usage(char *);

int main(int argc, char **argv)
{
    //////////////////////////////////////////////////////////////////
    //
    // Declare variables
    //
    char    string[25];
    char   *delim            = NULL;
    char   *str              = NULL;
    char    space            = ' ';
    char    unitflag         = 0;
    char    verbose          = 0;
    double  convert          = CONVERT;
    double  value            = 0.0;
    double  source           = 0.0;
    FILE   *inp              = NULL;
    FILE   *out              = stdout;
    uc      precis           = PRECIS;
    int     count            = -1;
    int     opt              = 0;

    //////////////////////////////////////////////////////////////////
    //
    // Argument processing loop
    //
    while ((opt = getopt(argc, argv, "C:d:hnNp:qT:vV")) != -1)
    {
        switch (opt)
        {
            case 'C':
                convert      = atof(optarg);
                break;

            case 'd':
                delim        = optarg;
                break;

            case 'h':
            case '?':
                usage(*argv);
                exit(0);

            case 'n':
                unitflag     = 1;
                break;

            case 'N':
                space        = '\b';
                break;

            case 'p':
                sscanf(optarg, "%[0123456789]", string);
                precis       = atoi(string);
                break;

            case 'q':
                out          = fopen("/dev/null", "r");
                break;

            case 'T':
                if (strcasecmp(optarg, "egyptian") == 0)
                    convert  = 7.700;
                else if (strcasecmp(optarg, "itinerary") == 0)
                    convert  = 10.250;
                else if (strcasecmp(optarg, "olympic") == 0)
                    convert  = 9.144;
                else if (strcasecmp(optarg, "persian") == 0)
                    convert  = CONVERT;
                else if (strcasecmp(optarg, "ptolemaic") == 0)
                    convert  = 8.699;
                else
                {
                    fprintf(stderr, "Invalid conversion type \"%s\"\n", optarg);
                    exit(1);
                }

                break;

            case 'v':
                verbose      = verbose + 1;
                if (verbose >  1)
                    count    = 0;
                break;

            case 'V':
                fprintf(stdout, "%s (%s), version %s\n", *(argv+0), NAME, VERSION);
                exit(0);

            default:
                exit(1);
        }
    }

    //////////////////////////////////////////////////////////////////
    //
    // If delimiter is not set, set it to default delimiter
    //
	if (delim               == NULL)
	{
		delim                = (char *) malloc (sizeof(char) * 5);
		strcat(delim, DELIM);
	}

    //////////////////////////////////////////////////////////////////
    //
    // Determine our source of input, STDIN or file
    //
    if (*(argv+optind)      == NULL)
        inp                  = stdin;
    else
    {
        inp                  = fopen(*(argv+optind), "r");
        if (inp             == NULL)
        {
            fprintf(stderr, "Error opening input file '%s'\n", *(argv+optind));
            exit(1);
        }
    }

    //////////////////////////////////////////////////////////////////
    //
    // If PRECIS environment variable is set, override precision
    //
    str                      = getenv("PRECIS");
    if (str                 != NULL)
        precis               = atoi(str);

    //////////////////////////////////////////////////////////////////
    //
    // Make some informational noise if verbosity is high enough
    //
    if (verbose             >  2)
    {
        fprintf(out, "Conversion rate set to %.*lf %s per %s.\n",
                     precis, convert, SRCUNITNAME, DSTUNITNAMESING);
    }

    //////////////////////////////////////////////////////////////////
    //
    // Input processing loop - keeping going until EOF is encountered
    //
    opt                      = -2;
    while (!feof(inp))
    {
        //////////////////////////////////////////////////////////////
        //
        // Grab (with input validation) a correct value; and break
        // out of the loop if we encounter an end to input
        //
        opt                  = fscanf(inp, "%[0123456789.]", string);
        if ((opt            == 0)    ||
            (opt            == EOF))
        {
            break;
        }
        fgetc(inp);

        //////////////////////////////////////////////////////////////
        //
        // Obtain source data and convert to destination value
        //
        source               = atof(string);
        value                = source / convert;

        //////////////////////////////////////////////////////////////
        //
        // Display final output, starting with a conditional display
        // of the source data, along with delimiter
        //
        if (verbose         >  0)
        {
            fprintf(out, "%.*lf", precis, source);
            if (unitflag    == 0)
            {
                if (space   == 32)
                    fprintf(out, " ");

                fprintf(out, "%s", SRCUNIT);
            }
			fprintf(out, "%s", delim);
        }

        //////////////////////////////////////////////////////////////
        //
        // Display destination value
        //
        fprintf(out, "%.*lf",   precis, value);
        if (unitflag        == 0)
        {
            if (space       == 32)
                fprintf(out, " ");
            fprintf(out, "%s", DSTUNIT);
        }
        fprintf(out, "\n");

        if (count           != -1)
            count            = count + 1;
    }

    //////////////////////////////////////////////////////////////////
    //
    // If enabled, display a count
    //
    if (count               != -1)
    {
        fprintf(out, "Processed %d item", count);
        if (count           != 1)
            fprintf(out, "s");
        fprintf(out, ".\n");
    }

    //////////////////////////////////////////////////////////////////
    //
    // Close custom file pointers
    //
    fclose(inp);
    fclose(out);

    return(0);
}

void usage(char *argv)
{
    fprintf(stdout, "Usage: %s [OPTION]... [FILE]\n", argv);
    fprintf(stdout, "Convert %s (%s) to %s (%s)\n\n", SRCUNITNAME, SRCUNIT, 
                                                      DSTUNITNAME, DSTUNIT);
    fprintf(stdout, "  -C VALUE    set custom conversion (%s per %s)\n",
                                                      SRCUNITNAME, DSTUNITNAME);
    fprintf(stdout, "  -d \"DELIM\"  set delimiter (default: \"%s\")\n", DELIM);
    fprintf(stdout, "  -h, -?      display this help and exit\n");
    fprintf(stdout, "  -n          do not display units\n");
    fprintf(stdout, "  -N          no space between value and unit\n");
    fprintf(stdout, "  -p POINTS   set output precision (default: %d)\n", PRECIS);
    fprintf(stdout, "  -q          quiet, do not display anything\n");
    fprintf(stdout, "  -T \"TYPE\"   specify cubit type (default: \"persian\")\n");
    fprintf(stdout, "  -v          verbosity level, display more information\n");
    fprintf(stdout, "  -V          display version information and exit\n\n");
    fprintf(stdout, "Verbosity compounds; first it will display source data,\n");
    fprintf(stdout, "next it will add a count of items processed. Then, it\n");
    fprintf(stdout, "will start reporting conversion factor information.\n\n");
    fprintf(stdout, "For TYPE specifier, choose from the following:\n");
    fprintf(stdout, "  egyptian    7.700%s\n", SRCUNIT);
    fprintf(stdout, "  itinerary   10.250%s\n", SRCUNIT);
    fprintf(stdout, "  olympic     9.144%s\n", SRCUNIT);
    fprintf(stdout, "  persian     %.3lf%s\n", CONVERT, SRCUNIT);
    fprintf(stdout, "  ptolemaic   8.699%s\n", SRCUNIT);
}
